/*
  adapter - Retrieve VPD for adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter.c,v 1.5 2006/03/08 01:22:58 martins Exp $
*/

#include <string.h>
#include <stdlib.h>

#include "lsvpd.h"
#include "init.h"
#include "vpd.h"
#include "util.h"

#include "adapter.h"

/********************************************************************/

/* Some of this could be done using pci.ids, but that would only work
 * for PCI adapters!  Also, the strings below tend to be comptabile
 * with AIX lsvpd. */

static struct {
	const char * subtype;
	const char * ds;
} adapter_ds_map[] = {
	{ ADAPTER_SUBTYPE_SCSI,       "SCSI I/O Controller"             },
	{ ADAPTER_SUBTYPE_ETHERNET,   "Ethernet PCI Adapter"            },
	{ ADAPTER_SUBTYPE_IDE,        "IDE I/O Controller"              },
	{ ADAPTER_SUBTYPE_USB,        "USB Host Controller"             },
	{ ADAPTER_SUBTYPE_DISPLAY,    "Display Adapter"                 },
	{ ADAPTER_SUBTYPE_SERIAL,     "Serial Adapter"                  },
	{ ADAPTER_SUBTYPE_SOUND,      "Multimedia Audio Controller"     },
	{ ADAPTER_SUBTYPE_FIBRE,      "Fibre Channel Adapter"           },
	{ ADAPTER_SUBTYPE_RAID,       "RAID Bus Controller"             },
	{ ADAPTER_SUBTYPE_SATA,       "Serial ATA Adapter"              },
	{ ADAPTER_SUBTYPE_ATA,        "ATA Adapter"                     },
	{ ADAPTER_SUBTYPE_ISCSI,      "iSCSI Adapter"                   },
	{ ADAPTER_SUBTYPE_SCSI_DEBUG, "SCSI Debug Fake Adapter"         },
	{ ADAPTER_SUBTYPE_IBMVETH,    "IBM Virtual Ethernet"            },
	{ ADAPTER_SUBTYPE_IBMVSCSI,   "Virtual SCSI I/O Controller"     },
	{ ADAPTER_SUBTYPE_OMS,        "Unknown Mass Storage Controller" },
	{ ADAPTER_SUBTYPE_TOKEN,      "Token Ring Network Controller"   },
	{ NULL,                       NULL                              },
};

static char *
adapter_get_ds_basic(const node_t type,
		     const bus_addr_t bus_addr,
		     const node_t subtype)
{
	char * ret;
	int i;
	const char * x;

	ret = NULL;;

	if (NULL != subtype) {
		i = 0;
		while ((NULL != (x = adapter_ds_map[i].subtype)) &&
		       (0 != strcmp(x, subtype)))
			i++;
		if (NULL != x)
			ret = strdup(adapter_ds_map[i].ds);
	}

	return ret;
}

adapter_get_ds_function_t *
adapter_get_ds_function = adapter_get_ds_basic;

char *
adapter_get_ds(const node_t type,
	       const bus_addr_t bus_addr,
	       const node_t subtype)
{
	char * ret = NULL;

	if (NULL != adapter_get_ds_function)
		ret = (*adapter_get_ds_function) (type, bus_addr, subtype);
	return ret;
}

/********************************************************************/

char *
adapter_get_source_node(const node_t bus_type,
			const bus_addr_t bus_addr,
			const node_t subtype)
{
	char * ret;
	node_t type;
	bus_info_t bus_info;

	ret = NULL;

	type = node_get_parent_subtype(bus_type, bus_addr);

	if (NULL != type) {
		bus_info = lsvpd_relative_filename(bus_type, bus_addr);
		if (NULL != bus_info) {
			ret = node_get_source_dir(type, bus_info, subtype);
			free(bus_info);
		}
		
		free(type);
	}

	return ret;
}

/********************************************************************/

static node_t
adapter_get_parent_subtype (const node_t     bus_type,
			    const bus_addr_t bus_addr)
{
	return strdup(bus_type);
}

/********************************************************************/

adapter_get_subtype_function_t * adapter_get_subtype_function = NULL;

node_t
adapter_get_subtype(const node_t type,
		    const bus_addr_t bus_addr)
{
	node_t ret = NULL;

	if (NULL != adapter_get_subtype_function)
		ret = (*adapter_get_subtype_function) (type, bus_addr);
	return ret;
}

/********************************************************************/

/* FIXME: This is temporary.  This implementation doesn't get all the
 * VPD, just the core fields.  Later this function will be moved or
 * will be changed to call (the equivalent of) adapter_extra_vpd_hook
 * and adapter_retrieve_logicals_vpd. This may result in the above
 * function becoming static. */

static bool
adapter_retrieve_vpd_default_partial(const node_t type,
				     const bus_info_t bus_info,
				     const char * node,
				     vpd_list_t ** vpd)
{
	bool ret;
	node_t bus_type;
	bus_addr_t bus_addr;
	node_t subtype;
	char * ds;

	ret = false;

	*vpd = NULL;

	if (bus_info_parse(bus_info, &bus_type, &bus_addr)) {
		subtype = adapter_get_subtype(bus_type, bus_addr);
		if (NULL != subtype) {
			ds = adapter_get_ds(bus_type, bus_addr, subtype);
			if ((NULL != ds) &&
			    (lsvpd_file_relative_write_string(node,
							      LSVPD_SUBTYPE_FILE,
							      subtype))) {
				*vpd = vpd_list_new();
				vpd_list_maybe_append_key_string(vpd, "DS", ds);
				free(ds);
			}
			free(subtype);
		}
		free(bus_type);
		free(bus_addr);
	}

	return NULL != *vpd;
}

/********************************************************************/

static const char * 
adapter_default_types[] = {
	ADAPTER_TYPE_PSEUDO,
	ADAPTER_TYPE_VIO,
	ADAPTER_TYPE_ISA,
	ADAPTER_TYPE_UNKNOWN,
	NULL,
};

static void
init(void)
{
	int i;

	node_functions_t a = {
		.type               = ADAPTER_TYPE_DEFAULT,
		.supertype          = NODE_TYPE_DEFAULT,
		.get_parent_subtype = adapter_get_parent_subtype,
		.retrieve_vpd       = adapter_retrieve_vpd_default_partial,
	};
	node_type_register(&a);

	/* Register other types, with appropriate supertype. */
	a.supertype = ADAPTER_TYPE_DEFAULT;
	a.get_parent_subtype = NULL;
	a.retrieve_vpd = NULL;
	i = 0;
	while (NULL != adapter_default_types[i]) {
		/* FIXME: Wacky const stuff happening here. */
		a.type = (node_t) adapter_default_types[i];
		node_type_register(&a);
		i++;
	}
}

INIT(init);
