/*
  adapter_device_tree - device-tree support routines for adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_device_tree.c,v 1.3 2006/03/08 04:54:32 martins Exp $
*/

/* For asprintf. */
#define _GNU_SOURCE

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "init.h"
#include "sysfs.h"
#include "device_tree.h"
#include "util.h"

#include "adapter_pci.h"
#include "adapter_pci_vpd_parse.h"
#include "adapter.h"

/********************************************************************/

struct {
	const char * field;
	const char * prop;
	const size_t expected;
	const int    start;
	const int    length;
} dt_pci_fields[] = {
	{ "vendor",            "vendor-id",           4, 2, 2 },
	{ "device",            "device-id",           4, 2, 2 },
	{ "subsystem_vendor",  "subsystem-vendor-id", 4, 2, 2 },
	{ "subsystem_device",  "subsystem-id",        4, 2, 2 },
	{ "class",             "class-code",          4, 1, 2 },
	{ NULL,                0, 0, 0 },
};

static char *
dt_get_field_value(const char * node,
		   const char * field)
{
	char * ret;
	int i;
	const char * x;
	unsigned char * buf;
	size_t len;

	ret = NULL;

	i = 0;
	while ((NULL != (x = dt_pci_fields[i].field)) &&
	       (0 != strcmp(x, field)))
		i++;
	if ((NULL != x) &&
	    (lsvpd_file_relative_read(node, dt_pci_fields[i].prop,
				      &buf, &len))) {
		if (dt_pci_fields[i].expected == len) {
			ret = lsvpd_hexify(&buf[dt_pci_fields[i].start],
					   dt_pci_fields[i].length);
			if (NULL != ret)
				lsvpd_strtolower(ret);
		}
		free(buf);
	}

	return ret;
}

/* This function assumes a single VPD chunk for the adapter.  This is
 * true for everything except system VPD, which is handled
 * elsewhere. */
static vpd_list_t *
dt_retrieve_ibm_vpd(const bus_addr_t bus_addr,
		    const node_t subtype,
		    const char * node)
{
	vpd_list_t * ret;
	char * source_node;
	unsigned char * buf;
	size_t len;
	uint32_t vl;
	int n;
	growbuf * vpd;

	ret = NULL;

	source_node = adapter_get_source_node(ADAPTER_TYPE_PCI,
					      bus_addr, subtype);

	if (NULL != source_node) {
		if (lsvpd_file_relative_read(source_node, "ibm,vpd",
					     &buf, &len)) {
			vl = 0;
			if (len > 4) {
				/* Construct 4 byte, BIG-endian length. */
				for (n = 0; n < 4; n++)
					vl = (vl << 8) + (uint32_t) buf[n];
				
				if (len >= 4 + vl) {
					vpd = growbuf_append(growbuf_new(vl),
							     &buf[4], vl);
					ret = adapter_pci_vpd_parse(vpd);
					growbuf_free(vpd);
				}
			}
			free(buf);
		}
		free(source_node);
	}

	return ret;
}

/********************************************************************/

static adapter_get_ds_function_t * dt_get_ds_basic = NULL;

static char *
dt_get_ds_function(const node_t bus_type,
		   const bus_addr_t bus_addr,
		   const node_t subtype)
{
	char * ret;
	char * source_node;
	char * ibm_name;
	char * t;
	char * w;

	ret = NULL;

	if (NULL != dt_get_ds_basic)
		ret = (*dt_get_ds_basic)(bus_type, bus_addr, subtype);

	source_node = adapter_get_source_node(bus_type, bus_addr, subtype);

	if (NULL != source_node) {
		ibm_name = 
			lsvpd_file_relative_read_string(source_node,
							"ibm,fw-adapter-name");
		if (0 == strcmp(subtype, ADAPTER_SUBTYPE_SCSI)) {
			if (NULL != ret) {
				if (NULL != ibm_name) {
					if (-1 != asprintf(&t, "%s %s",
							   ibm_name, ret)) {
						free(ret);
						ret = t;
					}
					free(ibm_name);
				}
					
				w = lsvpd_relative_filename(source_node,
							    "wide");
				if (NULL != w) {
					if ((0 == access(w, F_OK)) &&
					    (-1 != asprintf(&t, "Wide/%s",
							    ret))) {
						free(ret);
						ret = t;
					}
					free(w);
				}
			}
		} else if ((NULL != ibm_name) &&
			   ((0 == strcmp(subtype, ADAPTER_SUBTYPE_ETHERNET)) ||
			    (0 == strcmp(subtype, ADAPTER_SUBTYPE_IBMVETH)))) {
				free(ret);
				ret = ibm_name;
		}

		free(source_node);
	}

	return ret;
}

/********************************************************************/

static char *
dt_sysfs_adapter_get_source_dir(const node_t type,
				bus_info_t   bus_info,
				node_t       subtype)
{
	char * ret;
	node_t bus_type;
	bus_addr_t bus_addr;
	char * d;
	char * suffix;

	ret = NULL;

	if (bus_info_parse(bus_info, &bus_type, &bus_addr)) {
		if (-1 != asprintf(&d, "%s/bus/%s/devices/%s/devspec",
				   lsvpd_sysfs_dir, bus_type, bus_addr)) {
			suffix = lsvpd_file_read_string(d);
			if (NULL != suffix) {
				lsvpd_strnmove_trim(suffix,
						    (unsigned char *) suffix,
						    strlen(suffix));
				/* Avoid extra '/' in path. */
				if (-1 == asprintf(&ret, "%s%s",
						   lsvpd_dt_source_dir, suffix))
					ret = NULL;
				free(suffix);
			}
			free(d);
		}
		free(bus_type);
		free(bus_addr);
	}

	return ret;
}


static char *
dt_nosysfs_adapter_get_source_dir(const node_t type,
				  bus_info_t   bus_info,
				  node_t       subtype)
{
	char * ret;
	char * t;
	size_t len;

	ret = NULL;

	t = bus_alias_get(bus_info);

	if (NULL != t) {
		len = strlen(lsvpd_dt_db_dir);
		if ((0 == strncmp(t, lsvpd_dt_db_dir, len)) &&
		    ('/' == t[len]))
			ret = lsvpd_relative_filename(lsvpd_dt_source_dir,
						      &t[len+1]);
		free(t);
	}

	return ret;
}

node_t
dt_nosysfs_get_subtype(const node_t type,
		       const bus_addr_t bus_addr)
{
	return retrieve_subtype(type, bus_addr);
}

/********************************************************************/

static void
init(void)
{
	node_functions_t a = {
		.type           = ADAPTER_TYPE_DEFAULT,
	};
	char * dt_type;

	if (NULL != lsvpd_dt_source_dir) {
		adapter_pci_get_field_value_function = dt_get_field_value;

		/* Remember old function to call. */
		dt_get_ds_basic  = adapter_get_ds_function;
		adapter_get_ds_function = dt_get_ds_function;

		if (NULL != lsvpd_sysfs_dir) {
			a.get_source_dir = dt_sysfs_adapter_get_source_dir;
			node_type_register(&a);
		} else {
			a.get_source_dir = dt_nosysfs_adapter_get_source_dir;
			node_type_register(&a);
			adapter_get_subtype_function = dt_nosysfs_get_subtype;
		}

		/* If device-tree type is chrp, them look for ibm,vpd
		 * properties. */
		dt_type = lsvpd_file_relative_read_string(lsvpd_dt_source_dir,
							  "device_type");
		if (NULL != dt_type) {
			if (0 == strcmp(dt_type, "chrp"))
				adapter_pci_retrieve_vpd_first =
					dt_retrieve_ibm_vpd;
			free(dt_type);
		}
	}
}

INIT(init);
