/*
  adapter_sysfs - sysfs support routines for adapters.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: adapter_sysfs.c,v 1.2 2006/02/17 04:06:25 martins Exp $
*/

/* For asprintf. */
#define _GNU_SOURCE

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <glob.h>

#include "init.h"
#include "sysfs.h"
#include "util.h"

#include "adapter_pci.h"
#include "adapter.h"

/********************************************************************/

struct {
	const char * field;
	const size_t expected;
	const int    start;
	const int    length;
} sysfs_pci_fields[] = {
	{ "vendor",            7, 2, 4 },
	{ "device",            7, 2, 4 },
	{ "subsystem_vendor",  7, 2, 4 },
	{ "subsystem_device",  7, 2, 4 },
	{ "class",             9, 2, 4 },
	{ NULL,                0, 0, 0 },
};

static char *
sysfs_get_field_value(const char * node,
		      const char * field)
{
	char * ret;
	char * t;
	int i;
	const char * x;

	ret = NULL;

	t = lsvpd_file_relative_read_string(node, field);
	if (NULL != t) {
		i = 0;
		while ((NULL != (x = sysfs_pci_fields[i].field)) &&
		       (0 != strcmp(x, field)))
			i++;
		if (NULL != x) {
			if ((strlen(t) == sysfs_pci_fields[i].expected) &&
			    (sysfs_pci_fields[i].length <
			     sysfs_pci_fields[i].expected)) {
				memmove(t,
					t + sysfs_pci_fields[i].start,
					sysfs_pci_fields[i].length);
				t[sysfs_pci_fields[i].length] = '\0';
				ret = t;
			} else
				free(t);
		} else
			free(t);
	}

	return ret;
}

/********************************************************************/

char *
sysfs_get_config_file(const bus_addr_t bus_addr)
{
	char * ret;

	if (-1 == asprintf(&ret, "%s/bus/pci/devices/%s/config",
			   lsvpd_sysfs_dir, bus_addr))
		ret = NULL;

	return ret;
}

/********************************************************************/

struct {
	const char * driver;
	const char * subtype;
} sysfs_driver_subtypes[] = {
	{ "scsi_debug",    ADAPTER_SUBTYPE_SCSI_DEBUG },
	{ "ibmveth",       ADAPTER_SUBTYPE_IBMVETH    },
	{ "ibmvscsi",      ADAPTER_SUBTYPE_IBMVSCSI   },
	{ "ide-pmac",      ADAPTER_SUBTYPE_IDE        },
	{ "Promise IDE",   ADAPTER_SUBTYPE_IDE        },
	{ "Promise_IDE",   ADAPTER_SUBTYPE_IDE        },
	{ "sata_svw",      ADAPTER_SUBTYPE_SATA       },
	{ "ata_piix",      ADAPTER_SUBTYPE_SATA       },
	{ "pata_pdc2027x", ADAPTER_SUBTYPE_ATA        },
	{ "qla4xxx",       ADAPTER_SUBTYPE_ISCSI      },
	{ NULL,            NULL                       },
};

static char *
sysfs_adapter_get_driver(const node_t type,
			 const bus_addr_t bus_addr)
{
	char * ret;
	char * f;
	char * t;
	glob_t globbuf;

	ret = NULL;

	if (-1 != asprintf(&f, "%s/bus/%s/devices/%s/driver",
			   lsvpd_sysfs_dir, type, bus_addr)) {
		t = lsvpd_readlink(f);
		if (NULL != t) {
			ret = lsvpd_basename(t);
			free(t);
		}
		free(f);
	}

	/* "driver" link only exists after about * 2.6.10.  Try this
	 * instead: return the first driver subdirectory that has this
	 * adapter under it. */
	if ((NULL == ret) &&
	    (-1 != asprintf(&f, "%s/bus/%s/drivers/*/%s",
			    lsvpd_sysfs_dir, type, bus_addr))) {
		if ((0 == glob(f, 0, NULL, &globbuf)) &&
		    (globbuf.gl_pathc > 0)) {
			t = lsvpd_dirname(globbuf.gl_pathv[0]);
			if (NULL != t) {
				ret = lsvpd_basename(t);
				free(t);
			}
		}
		globfree(&globbuf);
		free(f);
	}

	return ret;
}

static node_t
sysfs_driver_get_subtype(const node_t type,
			 const bus_addr_t bus_addr)
{
	node_t ret;
	char * driver;
	int i;
	const char * x;

	ret = NULL;

	driver = sysfs_adapter_get_driver(type, bus_addr);
	if (NULL != driver) {
		i = 0;
		while ((NULL != (x = sysfs_driver_subtypes[i].driver)) &&
		       (0 != strcmp(x, driver)))
			i++;
		if (NULL != x)
			/* FIXME: can we change some types and avoid
			 * the strdup here? */
			ret = strdup(sysfs_driver_subtypes[i].subtype);

		free(driver);
	}

	return ret;
}

/* Use a case-insensitive comparison against pci_class here, just in case. */
struct {
	const char * pci_class;
	const char * subtype;
} sysfs_pci_class_subtypes[] = {
	{ "0180", ADAPTER_SUBTYPE_OMS      },
	{ "0100", ADAPTER_SUBTYPE_SCSI     },
	{ "0104", ADAPTER_SUBTYPE_RAID     },
	{ "0c04", ADAPTER_SUBTYPE_FIBRE    },
	{ "0200", ADAPTER_SUBTYPE_ETHERNET },
	{ "0201", ADAPTER_SUBTYPE_TOKEN    },
	{ "0101", ADAPTER_SUBTYPE_IDE      },
	{ "0c03", ADAPTER_SUBTYPE_USB      },
	{ "0300", ADAPTER_SUBTYPE_DISPLAY  },
	{ "0700", ADAPTER_SUBTYPE_SERIAL   },
	{ "0401", ADAPTER_SUBTYPE_SOUND    },
	{ "07",   ADAPTER_SUBTYPE_GENPCI   },
	{ NULL,   NULL                     },
};

static node_t
sysfs_pci_class_get_subtype(const node_t type,
			    const bus_addr_t bus_addr)
{
	node_t ret;
	char * source_node;
	char * pci_class;
	int i;
	const char * x;

	ret = NULL;

	source_node = adapter_get_source_node(type, bus_addr, NULL);
	if (NULL != source_node) {
		pci_class = adapter_pci_get_field_value(source_node, "class");
		if (NULL != pci_class) {
			i = 0;
			while ((NULL !=
				(x = sysfs_pci_class_subtypes[i].pci_class)) &&
			       /* This allows partial matches. */
			       (0 != strncasecmp(x, pci_class, strlen(x))))
				i++;
			if (NULL != x)
				ret = strdup(sysfs_pci_class_subtypes[i].subtype);
			free(pci_class);
		}
		free(source_node);
	}

	return ret;
}

node_t
sysfs_get_subtype(const node_t type,
		  const bus_addr_t bus_addr)
{
	node_t ret;

	ret = NULL;

	if ((0 == strcmp(BUS_TYPE_UNKNOWN, type)) &&
	    (0 == strncmp("ide", bus_addr, 3)))
		ret = strdup(ADAPTER_SUBTYPE_IDE);
	else 
		ret = sysfs_driver_get_subtype(type, bus_addr);

	if ((NULL == ret) && (0 == strcmp(ADAPTER_TYPE_PCI, type)))
		ret = sysfs_pci_class_get_subtype(type, bus_addr);

	return ret;
}

/********************************************************************/

static void
init(void)
{
	if (NULL != lsvpd_sysfs_dir) {
		adapter_pci_get_field_value_function = sysfs_get_field_value;
		adapter_pci_get_config_file_function = sysfs_get_config_file;
		adapter_get_subtype_function = sysfs_get_subtype;
	}
}

INIT(init);
