/*
  node_handler - Perform an increasing number of VPD-related functions.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: node_handler.c,v 1.11 2006/01/12 01:19:06 martins Exp $
*/

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "init.h"
#include "util.h"
#include "lsvpd.h"
#include "vpd.h"
#include "node.h"

/********************************************************************/

#define NODE_LIST              "list_nodes"
#define NODE_GET_VPD           "get_vpd"
#define NODE_PRINT_NAME        "get_name"
#define NODE_PRINT_KERNEL_NAME "get_kernel_name"
#define NODE_PRINT_PARENT_NAME "get_parent_name"

static void
usage(const char * prog)
{
	fprintf(stderr,
		"usage:"
		"\t%s %s\n"
		"\t%s %s type bus_info node\n"
		"\t%s %s type bus_info\n"
		"\t%s %s type bus_info\n"
		"\t%s %s type bus_info\n",
		prog, NODE_LIST,
		prog, NODE_GET_VPD,
		prog, NODE_PRINT_NAME,
		prog, NODE_PRINT_KERNEL_NAME,
		prog, NODE_PRINT_PARENT_NAME);
	fprintf(stderr, "\nEnvironment variable LSVPD_DB_DIR must be set.\n");
	exit(1);
}


static bool
node_get_vpd(const node_t     type,
	     const bus_info_t bus_info,
	     const char *     node)
{
	bool ret;
	vpd_list_t * vpd;
	vpd_list_list_t * vll;

	ret = false;

	vpd = NULL;
	vll = NULL;

	ret = node_retrieve_vpd(type, bus_info, node, &vpd) &&
		(NULL != vpd) &&
		((vll = vpd_list_list_new()),
		 (NULL != (vll = vpd_list_list_append(vll, vpd)))) &&
		vpd_list_list_to_node(vll, node);

	if (NULL != vll)
		vpd_list_list_free(vll);
	else if (NULL != vpd)
		vpd_list_free(vpd);
	
	return ret;
}

bool node_list(void)
{
	bool ret;
	bus_info_list_t * l;
	bus_info_list_t * t;

	lsvpd_db_mkdirs();

	l = NULL;
	ret = node_list_devices(&l);

	if (ret)
		for (t = l; NULL != t; t = t->next)
			printf("%s\n", t->bus_info);

	node_list_clear(&l);

	return ret;
}

bool node_print_name(const node_t     type,
		     const bus_info_t bus_info)
{
	bool ret;
	char * n;

	ret = false;

	n = node_get_name(type, bus_info);
	if (NULL != n) {
		ret = true;
		printf("%s\n", n);
		free(n);
	}

	return ret;
}

bool node_print_kernel_name(const node_t     type,
			    const bus_info_t bus_info)
{
	bool ret;
	char * n;

	ret = false;

	n = node_get_kernel_name(type, bus_info);
	if (NULL != n) {
		ret = true;
		printf("%s\n", n);
		free(n);
	}

	return ret;
}

bool node_print_parent_name(const node_t     type,
			    const bus_info_t bus_info)
{
	bool ret;
	char * n;

	ret = false;

	n = node_get_parent_name(type, bus_info);
	if (NULL != n) {
		ret = true;
		printf("%s\n", n);
		free(n);
	}

	return ret;
}

int
main(int argc, char ** argv)
{
	bool ret;

	ret = false;

	 /*
	  * Placing the call(s) to call_inits() is an interesting
	  * problem.  If it appears before the following if statement,
	  * then checks in the init functions can stop us making it
	  * through to the usage() message below.  Putting calls to
	  * call_inits() in the above functions means they get buried.
	  * Repeating them here, before each function call makes them
	  * pretty obvious.
	  *
	  * Also, various init() functions dynamically allocate memory
	  * that is not freed.  Since this only happens once, this is
	  * OK.
	  */

	if ((2 == argc) &&
	    (0 == strcmp(argv[1], NODE_LIST))) {
		call_inits();
		ret = node_list();
	} else if ((5 == argc) &&
		   (0 == strcmp(argv[1], NODE_GET_VPD))) {
		call_inits();
		ret = node_get_vpd(argv[2], argv[3], argv[4]);; 
	} else if ((4 == argc) &&
		   (0 == strcmp(argv[1], NODE_PRINT_NAME))) {
		call_inits();
		ret = node_print_name(argv[2], argv[3]); 
	} else if ((4 == argc) &&
		   (0 == strcmp(argv[1], NODE_PRINT_KERNEL_NAME))) {
		call_inits();
		ret = node_print_kernel_name(argv[2], argv[3]); 
	} else if ((4 == argc) &&
		   (0 == strcmp(argv[1], NODE_PRINT_PARENT_NAME))) {
		call_inits();
		ret = node_print_parent_name(argv[2], argv[3]); 
	} else
		usage(argv[0]);

	return ret ? 0 : 1;
}
